/*
 *   $Id: slist.c,v 1.19 2002/11/29 22:23:18 ondrap Exp $
 *
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999,2000,2001,2002 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>

#include "segment.h"
#include "konvert.h"
#include "stack.h"
#include "mathem.h"
#include "slist.h"
#include "MathLib.h"
#include "complex.h"
#include "display.h"
#include "funcs.h"
#include "matrix.h"
#include "cmatrix.h"

void 
list_delete(List *lst) 
{
	MemPtrFree(lst);
}


/***********************************************************************
 *
 * FUNCTION:     list_new
 * 
 * DESCRIPTION:  Allocate new structure for the list
 *
 * PARAMETERS:   size (number of items)
 *
 * RETURN:       New list
 *      
 ***********************************************************************/
List *
list_new(Int16 size)
{
	List *result;

	result = MemPtrNew(sizeof(*result) + size*sizeof(result->item[0]));
	if (result == NULL)
	    return NULL;

	result->size = size;

	return result;
}

/***********************************************************************
 *
 * FUNCTION:     list_dup
 * 
 * DESCRIPTION:  Duplicate list (allocate new space, copy..)
 *
 * PARAMETERS:   list
 *
 * RETURN:       duplicated list
 *      
 ***********************************************************************/
List *
list_dup(List *oldlist)
{
	List *result;
	Int16 i;
	
	result = list_new(oldlist->size);
	if (!result)
	    return NULL;
	for (i=0;i<oldlist->size;i++)
		result->item[i] = oldlist->item[i];
	return result;
}


/***********************************************************************
 *
 * FUNCTION:     list_sort
 * 
 * DESCRIPTION:  Sort a list, this functions sorts the numbers according
 *               their absolute value but it respects sign of real part
 *
 * PARAMETERS:   lst - list
 *
 * RETURN:       sorted list
 *      
 ***********************************************************************/
#define CP_ABS(x) (hypot(x.real,x.imag)*sgn(x.real))

static void list_sort(List *lst) NEWFUNC;
static void
list_sort(List *lst)
{
	Int16 gap,i,j;
	Complex temp;

	for (gap = lst->size/2;gap > 0; gap/=2)
		for (i = gap;i < lst->size;i++) {
			for (j=i-gap;j>=0
			  && CP_ABS(lst->item[j]) > CP_ABS(lst->item[j+gap]);
			     j-=gap) {
				temp = lst->item[j];
				lst->item[j] = lst->item[j+gap];
				lst->item[j+gap] = temp;
			}
		}
}

/***********************************************************************
 *
 * FUNCTION:     list_real
 * 
 * DESCRIPTION:  Test if list contains only real numbers
 *
 * PARAMETERS:   list
 *
 * RETURN:       true - only real numbers
 *               false - contains complex numbers
 *      
 ***********************************************************************/
static Boolean list_real(List *lst) NEWFUNC;
static Boolean
list_real(List *lst)
{
	Int16 i;
	
	for (i=0;i < lst->size;i++)
		if (lst->item[i].imag != 0.0)
			return false;

	return true;
}

/***********************************************************************
 *
 * FUNCTION:     list_func2
 * 
 * DESCRIPTION:  Function that modify lists (sorting etc.)
 *
 * PARAMETERS:   list on stack
 *
 * RETURN:       list on stack
 *      
 ***********************************************************************/
CError
list_func2(Functype *func,CodeStack *stack)
{
	List *lst;
	CError err;
	Int16 i;
	Complex tmp;

	err = stack_get_val(stack,&lst,list);
	if (err)
		return err;

	switch (func->num) {
	case LIST_SORTA:
		list_sort(lst);
		break;
	case LIST_SORTD:
		list_sort(lst);
		for (i=0;i < lst->size/2;i++) {
			tmp = lst->item[i];
			lst->item[i] = lst->item[lst->size - i - 1];
			lst->item[lst->size - i - 1] = tmp;
		}
		break;
	case LIST_CUMSUM:
		for (i=1;i<lst->size;i++) {
			lst->item[i].real += lst->item[i-1].real;
			lst->item[i].imag += lst->item[i-1].imag;
		}
		break;
	}
	stack_add_val(stack,&lst,list);
	list_delete(lst);
	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     list_func
 * 
 * DESCRIPTION:  Functions work on top of complex lists and return
 *               complex numbers
 *
 * PARAMETERS:   list on stack
 *
 * RETURN:       complex on stack
 *      
 ***********************************************************************/
CError
list_func(Functype *func,CodeStack *stack)
{
	List *lst;
	CError err;
	Complex result;
	Int16 i;

	err = stack_get_val(stack,&lst,list);
	if (err)
		return err;

	switch (func->num) {
	case LIST_DIM:
		result.real = lst->size;
		result.imag = 0.0;
		break;
	case LIST_MIN:
		list_sort(lst);
		result = lst->item[0];
		break;
	case LIST_MAX:
		list_sort(lst);
		result = lst->item[lst->size - 1];
		break;
	case LIST_SUM:
		result.real = 0.0;
		result.imag = 0.0;
		for (i=0;i<lst->size;i++)
			result = cplx_add(result,lst->item[i]);
		break;
	case LIST_PROD:
		result.real = lst->item[0].real;
		result.imag = lst->item[0].imag;
		for (i=1;i < lst->size;i++)
			result = cplx_multiply(result,lst->item[i]);
		break;
	case LIST_MEDIAN:
		list_sort(lst);
		if (lst->size % 2)
			result = lst->item[lst->size/2];
		else {
			result = cplx_add(lst->item[lst->size/2-1],
					  lst->item[lst->size/2]);
			result.real /= 2;
			result.imag /= 2;
		}
		break;
	}

	if (result.imag != 0.0)
		stack_add_val(stack,&result,complex);
	else
		stack_add_val(stack,&result.real,real);

	list_delete(lst);
	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     list_sum
 * 
 * DESCRIPTION:  Compute a total sum of a list
 *
 * PARAMETERS:   lst - list
 *
 * RETURN:       sum(lst)
 *      
 ***********************************************************************/
static double list_sum(List *lst) NEWFUNC;
static double
list_sum(List *lst)
{
	Int16 i;
	double result = 0.0;
	
	for (i=0;i<lst->size;i++)
		result += lst->item[i].real;
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     list_mean
 * 
 * DESCRIPTION:  Compute a mean of a list
 *
 * PARAMETERS:   lst - list
 *
 * RETURN:       mean(lst)
 *      
 ***********************************************************************/
static double list_mean(List *lst) NEWFUNC;
static double
list_mean(List *lst)
{
	Int16 i;
	double result;

	result = 0.0;
	for (i=0;i<lst->size;i++)
		result += lst->item[i].real;
	result /= lst->size;
	
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     list_moment
 * 
 * DESCRIPTION:  Return nth moment of the list
 *
 * PARAMETERS:   list 
 *               moment
 *
 * RETURN:       result
 *      
 ***********************************************************************/
static double list_moment(List *lst,Int16 moment) NEWFUNC;
static double
list_moment(List *lst,Int16 moment)
{
	Int16 i,j;
	double result,tmp,mean;
	
	mean = list_mean(lst);
	result = 0.0;
	for (i=0;i<lst->size;i++) {
		tmp = (lst->item[i].real - mean);
		for (j=1;j<moment;j++)
			tmp *= (lst->item[i].real - mean);
		result += tmp;
	}
	result /= lst->size;

	return result;
}

/***********************************************************************
 *
 * FUNCTION:     list_variance
 * 
 * DESCRIPTION:  Compute variance of a list
 *
 * PARAMETERS:   List
 *
 * RETURN:       sum((x-mean(x))^2)/(n-1)
 *      
 ***********************************************************************/
static double list_variance(List *lst) NEWFUNC;
static double 
list_variance(List *lst)
{
	double mean;
	double result;
	Int16 i;

	if (lst->size == 1)
		return 0.0;

	mean = list_mean(lst);
	result = 0.0;
	for (i=0;i<lst->size;i++)
		result += (mean - lst->item[i].real)*(mean-lst->item[i].real);
	result /= lst->size - 1;

	return result;
}

static char * list_linreg(List *lstx,List *lsty) NEWFUNC;
static char *
list_linreg(List *lstx,List *lsty)
{
	double b1,b0;
	double tmp,tmp2;
	Int16 i;
	double n;
	char *result,*res1,*res2;

	n = lstx->size; /* = lsty->size */

	for (i=0,tmp=0.0,tmp2=0.0;i<n;i++) {
		tmp += lstx->item[i].real * lsty->item[i].real;
		tmp2 += lstx->item[i].real * lstx->item[i].real;
	}
	tmp *= n;
	tmp -= list_sum(lstx) * list_sum(lsty);
	tmp2 *= n;
	tmp2 -= list_sum(lstx)*list_sum(lstx);
	b1 = tmp / tmp2;
	b0 = list_sum(lsty)/n - b1*list_sum(lstx)/n;
	
	res1 = display_real(b0);
	res2 = display_real(b1);
	result = MemPtrNew(StrLen(res1)+StrLen(res2)+10);
	StrCopy(result,res2);
	StrCat(result,"x+");
	StrCat(result,res1);

	MemPtrFree(res1);
	MemPtrFree(res2);
	
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     list_mode
 * 
 * DESCRIPTION:  Find a modus (most frequent value) in a list
 *
 * PARAMETERS:   list
 *
 * RETURN:       most frequent value
 *      
 ***********************************************************************/
static double list_mode(List *lst) NEWFUNC;
static double
list_mode(List *lst)
{
	double last, most;
	Int16 mostcount,count,i;

	list_sort(lst);

	last = most = lst->item[0].real;
	mostcount = count = 1;
	
	for (i=1;i<lst->size;i++) {
		if (lst->item[i].real != last) {
			last = lst->item[i].real;
			count = 0;
		}
		count ++;
		if (count > mostcount) {
			mostcount = count;
			most = last;
		}
	}
	return most;
}

CError 
list_regr(Functype *func,CodeStack *stack)
{
	List *lstx,*lsty;
	CError err;
	char *result;

	err = stack_get_val2(stack,&lstx,&lsty,list);
	if (err)
		return err;
	
	if (!list_real(lstx) || !list_real(lsty) ||
	    lstx->size != lsty->size)
		return c_badarg;

	switch (func->num) {
	case LIST_LINREG:
		result = list_linreg(lstx,lsty);
		break;
	default:
		return c_internal;
	}
	
	stack_add_val(stack,result,string);
	MemPtrFree(result);
	list_delete(lstx);
	list_delete(lsty);

	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     list_stat
 * 
 * DESCRIPTION:  Basic statistical functions that work with
 *               'real' lists and return a real number
 *               Some functions may be modified by a second parameter,
 *               that if non-zero forces the variation to be defined
 *               as 'sum((x-mean(x))^2)/n', 
 *               if zero: 'sum((x-mean(x))^2)/(n-1)'
 *
 * PARAMETERS:   list on stack, optional int
 *
 * RETURN:       real num on stack
 *      
 ***********************************************************************/
CError
list_stat(Functype *func,CodeStack *stack)
{
	List *lst;
	CError err;
	double result;
	UInt32 intarg = 0;
	Int16 i;

	if (func->paramcount > 2) 
		return c_badargcount;
	else if (func->paramcount == 2) {
		err = stack_get_val(stack,&intarg,integer);
		if (err)
			return err;
	}

	err = stack_get_val(stack,&lst,list);
	if (err)
		return err;
	
	if (!list_real(lst)) {
		list_delete(lst);
		return c_badarg;
	}

	switch (func->num) {
	case LIST_SKEWNESS:
		if (list_moment(lst,2) != 0.0) {
			result = list_moment(lst,3);
			result /= pow(sqrt(list_moment(lst,2)),3);
		} else
			result = 0.0;
		break;
	case LIST_KURTOSIS:
		if (list_moment(lst,2) != 0.0) {
			result = list_moment(lst,4);
			result /= pow(list_moment(lst,2),2);
			result -= 3.0;
		} else
			result = 0.0;
		break;
	case LIST_MOMENT:
		if (intarg == 0) {
			list_delete(lst);
			return c_badarg;
		}
		result = list_moment(lst,intarg);
		break;
	case LIST_MEAN:
		result = list_mean(lst);
		break;
	case LIST_GMEAN:
		for (i=0,result=1.0;i<lst->size;i++)
			result *= lst->item[i].real;
		result = pow(result,1.0/((double) lst->size));
		break;
	case LIST_VARIANCE:
		if (intarg)
			result = list_moment(lst,2);
		else
			result = list_variance(lst);
		break;
	case LIST_STDDEV:
		if (intarg)
			result = sqrt(list_moment(lst,2));
		else
			result = sqrt(list_variance(lst));
		break;
	case LIST_MODE:
		result = list_mode(lst);
		break;
	case LIST_VARCOEF:
		result = sqrt(list_variance(lst))/list_mean(lst);
		break;
	}
	stack_add_val(stack,&result,real);
	list_delete(lst);

	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     list_input
 * 
 * DESCRIPTION:  The list() function, takes all arguments and 
 *               creates a list
 *               If the argument is list, matrix is assumed and
 *               matrix is built
 *
 * PARAMETERS:   1..n parametrs on stack
 *
 * RETURN:       list on stack
 *      
 ***********************************************************************/
CError
list_input(Functype *func,CodeStack *stack)
{
	List *l = NULL;
	Matrix *m = NULL;
	CMatrix *cm = NULL;
	Int16 i,j;
	CError err;
	rpntype itemtype;
	Int16 rows,cols=0;

	/* First get the type of the first parameter */
	err = stack_item_type(stack,&itemtype,0);
	if (err)
		return err;
	
	if (itemtype != list) {
		/* We will build a list now */
		l = list_new(func->paramcount);
		for (i=func->paramcount-1;i>=0;i--) {
			err = stack_get_val(stack,&l->item[i],complex);
			if (err) {
				list_delete(l);
			return err;
			}
		}
		stack_add_val(stack,&l,list);
		list_delete(l);
		return c_noerror;
	}
	/* We will build matrix, because we got list as argument */
        /* First we build complex matrix and then try to scale it down */
	rows = func->paramcount;
	for (i=func->paramcount-1;i>=0;i--) {
		err = stack_get_val(stack,&l,list);
		if (err) {
			if (cm) cmatrix_delete(cm);
			return err;
		}
		if (!cm) {
			/* The first argument must have correct matrix size */
			cols = l->size;
			cm = cmatrix_new(rows,cols);
		}
		if (l->size > cols) {
			cmatrix_delete(cm);
			list_delete(l);
			return c_baddim;
		}
		for (j=0;j < l->size;j++) 
			MATRIX(cm,i,j) = l->item[j];
		
		list_delete(l);
	}
	/* Try to scale the matrix down to normal one */
	m = cmatrix_to_matrix(cm);
	if (m) {
		stack_add_val(stack,&m,matrix);
		matrix_delete(m);
	} else 
		stack_add_val(stack,&cm,cmatrix);
	cmatrix_delete(cm);
	return c_noerror;
}


/***********************************************************************
 *
 * FUNCTION:     list_rand
 * 
 * DESCRIPTION:  Create an array of random numbers (or randNorm)
 *
 * PARAMETERS:   size of the list
 *
 * RETURN:       list of random numbers
 *      
 ***********************************************************************/
CError
list_rand(Functype *func,CodeStack *stack)
{
	List *result;
	CError err;
	UInt32 intarg;
	UInt32 i;

	err = stack_get_val(stack,&intarg,integer);
	if (err)
		return err;
	
	if (intarg == 0)
		return c_badarg;

	result = list_new(intarg);
	if (!result)
	    return c_memory;

	switch (func->num) {
	case LIST_RAND:
		for (i=0;i < intarg; i++) {
			result->item[i].real = RAND_NUM;
			result->item[i].imag = 0.0;
		}
		break;
	case LIST_RNORM:
	default:
		for (i=0;i < intarg; i++) {
			result->item[i].imag = 0.0;
			result->item[i].real = 2*sqrt(-0.5*log(RAND_NUM))*cos(RAND_NUM*2*M_PIl);
		}
		break;
	}
	err = stack_add_val(stack,&result,list);
	list_delete(result);

	return err;
}

/***********************************************************************
 *
 * FUNCTION:     list_item
 * 
 * DESCRIPTION:  The operation of the list[2] type
 *
 * PARAMETERS:   list, item
 *
 * RETURN:       list[item]
 *      
 ***********************************************************************/
CError
list_item(Functype *func, CodeStack *stack)
{
	List *lst;
	CError err;
	UInt32 intarg;
	Complex result;

	err = stack_get_val(stack,&intarg,integer);
	if (err)
		return err;

	err = stack_get_val(stack,&lst,list);
	if (err)
		return err;
	
	if (intarg == 0 || intarg > lst->size)
		return c_baddim;
	
	result = lst->item[intarg-1];
	stack_add_val(stack,&result,complex);
	
	list_delete(lst);
	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     list_range
 * 
 * DESCRIPTION:  range(num[:start[:step]]) generator
 *
 * PARAMETERS:   num[:start[:step]]
 *
 * RETURN:       list[item]
 *      
 ***********************************************************************/CError
list_range(Functype *func, CodeStack *stack)
{
	List *result;
	CError err;
	UInt32 listlen;
	UInt32 i;
	double first = 1.0;
	double step = 1.0;

	if (func->paramcount > 3)
		return c_badargcount;
	if (func->paramcount == 3) {
		err = stack_get_val(stack,&step,real);
		if (err)
			return err;
	}
	if (func->paramcount >= 2) {
		err = stack_get_val(stack,&first,real);
		if (err)
			return err;
	}
	
	err = stack_get_val(stack,&listlen,integer);
	if (err)
		return err;
	if (listlen == 0)
		return c_badarg;
	

	result = list_new(listlen);
	for (i=0;i < listlen;i++) {
		result->item[i].real = first + i*step;
		result->item[i].imag = 0.0;
	}
	stack_add_val(stack,&result,list);
	list_delete(result);

	return c_noerror;
}

CError
list_mathem(Functype *func, CodeStack *stack)
{
	double arg1,arg2,res;
	List *result;
	CError err;
	
	if ((err=stack_get_val2(stack,&arg1,&arg2,real)))
		return err;
	
	if (func->num==MATH_PTORX || func->num==MATH_PTORY) 
		arg2 = math_user_to_rad(arg2);
	
	result = list_new(2);
	result->item[0].imag = 0.0;
	result->item[1].imag = 0.0;

	switch (func->num) {
	case LIST_RTOP:
		res = atan2(arg2,arg1);
		res = math_rad_to_user(res);
		result->item[0].real = hypot(arg1,arg2);
		result->item[1].real = res;
		break;
	case LIST_PTOR:
		arg2 = math_user_to_rad(arg2);
		result->item[0].real = cos(arg2)*arg1;
		result->item[1].real = sin(arg2)*arg1;
		break;
	}
	stack_add_val(stack,&result,list);
	list_delete(result);
	
	return c_noerror;	
}
